@@ -23,6 +23,7 @@
 # OTHER DEALINGS IN THE SOFTWARE.
 
 import io
+import time
 import codecs
 import contextlib
 import functools
@@ -73,6 +74,7 @@ from datetime import timedelta
 from errno import EAGAIN
 
 old_socket = socket.socket
+old_socketpair = getattr(socket, 'socketpair', None)
 old_SocketType = socket.SocketType
 old_create_connection = socket.create_connection
 old_gethostbyname = socket.gethostbyname
@@ -146,8 +148,7 @@ def FALLBACK_FUNCTION(x):
 
 
 class HTTPrettyRequest(BaseHTTPRequestHandler, BaseClass):
-    r"""
-    Represents a HTTP request. It takes a valid multi-line,
+    r"""Represents a HTTP request. It takes a valid multi-line,
     ``\r\n`` separated string with HTTP headers and parse them out using
     the internal `parse_request` method.
 
@@ -160,15 +161,24 @@ class HTTPrettyRequest(BaseHTTPRequestHandler, BaseClass):
     ``headers`` -> a mimetype object that can be cast into a dictionary,
     contains all the request headers
 
-    ``method`` -> the HTTP method used in this request
+    ``protocol`` -> the protocol of this host, inferred from the port
+    of the underlying fake TCP socket.
+
+    ``host`` -> the hostname of this request.
+
+    ``url`` -> the full url of this request.
+
+    ``path`` -> the path of the request.
+
+    ``method`` -> the HTTP method used in this request.
 
     ``querystring`` -> a dictionary containing lists with the
     attributes. Please notice that if you need a single value from a
     query string you will need to get it manually like:
 
-    ``body`` -> the request body as a string
+    ``body`` -> the request body as a string.
 
-    ``parsed_body`` -> the request body parsed by ``parse_request_body``
+    ``parsed_body`` -> the request body parsed by ``parse_request_body``.
 
     .. testcode::
 
@@ -176,16 +186,15 @@ class HTTPrettyRequest(BaseHTTPRequestHandler, BaseClass):
       {'name': ['Gabriel Falcao']}
       >>> print request.querystring['name'][0]
 
-
-
     """
-    def __init__(self, headers, body=''):
+    def __init__(self, headers, body='', sock=None, path_encoding = 'iso-8859-1'):
         # first of all, lets make sure that if headers or body are
         # unicode strings, it must be converted into a utf-8 encoded
         # byte string
+        self.created_at = time.time()
         self.raw_headers = utf8(headers.strip())
         self._body = utf8(body)
-
+        self.connection = sock
         # Now let's concatenate the headers with the body, and create
         # `rfile` based on it
         self.rfile = io.BytesIO(b'\r\n\r\n'.join([self.raw_headers, self.body]))
@@ -205,14 +214,13 @@ class HTTPrettyRequest(BaseHTTPRequestHandler, BaseClass):
         if not self.parse_request():
             return
 
-        # making the HTTP method string available as the command
-        self.method = self.command
-
         # Now 2 convenient attributes for the HTTPretty API:
 
-        # `querystring` holds a dictionary with the parsed query string
+        # - `path`
+        # - `querystring` holds a dictionary with the parsed query string
+        # - `parsed_body` a string
         try:
-            self.path = self.path.encode('iso-8859-1')
+            self.path = self.path.encode(path_encoding)
         except UnicodeDecodeError:
             pass
 
@@ -231,6 +239,25 @@ class HTTPrettyRequest(BaseHTTPRequestHandler, BaseClass):
         """
         self.parsed_body = self.parse_request_body(self._body)
 
+    @property
+    def method(self):
+        """the HTTP method used in this request"""
+        return self.command
+
+    @property
+    def protocol(self):
+        """the protocol used in this request"""
+        proto = ''
+        if not self.connection:
+            return ''
+        elif self.connection.is_http:
+            proto = 'http'
+
+        if self.connection.is_secure:
+            proto = 'https'
+
+        return proto
+
     @property
     def body(self):
         return self._body
@@ -246,11 +273,21 @@ class HTTPrettyRequest(BaseHTTPRequestHandler, BaseClass):
     def __nonzero__(self):
         return bool(self.body) or bool(self.raw_headers)
 
+    @property
+    def url(self):
+        """the full url of this recorded request"""
+        return "{}://{}{}".format(self.protocol, self.host, self.path)
+
+    @property
+    def host(self):
+        return self.headers.get('Host') or '<unknown>'
+
     def __str__(self):
-        tmpl = '<HTTPrettyRequest("{}", total_headers={}, body_length={})>'
+        tmpl = '<HTTPrettyRequest("{}", "{}", headers={}, body={})>'
         return tmpl.format(
-            self.headers.get('content-type', ''),
-            len(self.headers),
+            self.method,
+            self.url,
+            dict(self.headers),
             len(self.body),
         )
 
@@ -360,6 +397,10 @@ class FakeAddressTuple(object):
         raise AssertionError('socket {} is not connected'.format(self.fakesocket.truesock))
 
 
+def fake_socketpair(*args, **kw):
+    with restored_libs():
+        return old_socketpair(*args, **kw)
+
 class fakesock(object):
     """
     fake :py:mod:`socket`
@@ -370,7 +411,7 @@ class fakesock(object):
         _entry = None
         debuglevel = 0
         _sent_data = []
-
+        is_secure = False
         def __init__(
             self,
             family=socket.AF_INET,
@@ -395,7 +436,14 @@ class fakesock(object):
             self.is_http = False
             self._bufsize = 32 * 1024
 
-        def create_socket(self):
+        def __repr__(self):
+            return '{self.__class__.__module__}.{self.__class__.__name__}("{self.host}")'.format(**locals())
+
+        @property
+        def host(self):
+            return ":".join(map(str, self._address))
+
+        def create_socket(self, address=None):
             return old_socket(self.socket_family, self.socket_type, self.socket_proto)
 
         def getpeercert(self, *a, **kw):
@@ -447,14 +495,15 @@ class fakesock(object):
                 ports_to_check = (
                     POTENTIAL_HTTP_PORTS.union(POTENTIAL_HTTPS_PORTS))
                 self.is_http = self._port in ports_to_check
+                self.is_secure = self._port in POTENTIAL_HTTPS_PORTS
 
             if not self.is_http:
-                self.connect_truesock()
-            elif self.truesock and not self.__truesock_is_connected__:
+                self.connect_truesock(address=address)
+            elif self.truesock and not self.real_socket_is_connected():
                 # TODO: remove nested if
                 matcher = httpretty.match_http_address(self._host, self._port)
                 if matcher is None:
-                    self.connect_truesock()
+                    self.connect_truesock(address=address)
 
         def bind(self, address):
             self._address = (self._host, self._port) = address
@@ -465,19 +514,30 @@ class fakesock(object):
             if httpretty.allow_net_connect and not self.truesock:
                 self.truesock = self.create_socket()
             elif not self.truesock:
-                raise UnmockedError()
+                raise UnmockedError('Failed to socket.bind() because because a real socket was never created.', address=address)
 
             return self.truesock.bind(address)
 
-        def connect_truesock(self):
-            if httpretty.allow_net_connect and not self.truesock:
-                self.truesock = self.create_socket()
-            elif not self.truesock:
-                raise UnmockedError()
+        def connect_truesock(self, request=None, address=None):
+            address = address or self._address
 
             if self.__truesock_is_connected__:
                 return self.truesock
-            with restored_libs():
+
+            if request:
+                logger.warning('real call to socket.connect() for {request}'.format(**locals()))
+            elif address:
+                logger.warning('real call to socket.connect() for {address}'.format(**locals()))
+            else:
+                logger.warning('real call to socket.connect()')
+
+            if httpretty.allow_net_connect and not self.truesock:
+                self.truesock = self.create_socket(address)
+            elif not self.truesock:
+                raise UnmockedError('Failed to socket.connect() because because a real socket was never created.', request=request, address=address)
+
+            undo_patch_socket()
+            try:
                 hostname = self._address[0]
                 port = 80
                 if len(self._address) == 2:
@@ -490,6 +550,9 @@ class fakesock(object):
                 sock.connect(self._address)
                 self.__truesock_is_connected__ = True
                 self.truesock = sock
+            finally:
+                apply_patch_socket()
+
             return self.truesock
 
         def real_socket_is_connected(self):
@@ -540,17 +603,22 @@ class fakesock(object):
             buffer so that HTTPretty can return it accordingly when
             necessary.
             """
+            request = kw.pop('request', None)
+            if request:
+                bytecount = len(data)
+                logger.warning('{self}.real_sendall({bytecount} bytes) to {request.url} via {request.method} at {request.created_at}'.format(**locals()))
 
             if httpretty.allow_net_connect and not self.truesock:
-                self.connect_truesock()
+
+                self.connect_truesock(request=request)
             elif not self.truesock:
-                raise UnmockedError()
+                raise UnmockedError(request=request)
 
             if not self.is_http:
                 self.truesock.setblocking(1)
                 return self.truesock.sendall(data, *args, **kw)
 
-            sock = self.connect_truesock()
+            sock = self.connect_truesock(request=request)
 
             sock.setblocking(1)
             sock.sendall(data, *args, **kw)
@@ -612,7 +680,7 @@ class fakesock(object):
                     else:
                         self._entry.request.body += body
 
-                    httpretty.historify_request(headers, body, False)
+                    httpretty.historify_request(headers, body, sock=self)
                     return
 
             if path[:2] == '//':
@@ -627,7 +695,7 @@ class fakesock(object):
                 headers = ''
                 body = data
 
-            request = httpretty.historify_request(headers, body)
+            request = httpretty.historify_request(headers, body, sock=self)
 
             info = URIInfo(
                 hostname=self._host,
@@ -641,26 +709,14 @@ class fakesock(object):
 
             if not entries:
                 self._entry = None
-                self.real_sendall(data)
+                self.real_sendall(data, request=request)
                 return
 
             self._entry = matcher.get_next_entry(method, info, request)
 
         def forward_and_trace(self, function_name, *a, **kw):
-            if self.truesock and not self.__truesock_is_connected__:
-                self.truesock = self.create_socket()
-                ### self.connect_truesock()
-
-            if self.__truesock_is_connected__:
-                function = getattr(self.truesock, function_name)
-
-            if self.is_http:
-                if self.truesock and not self.__truesock_is_connected__:
-                    self.truesock = self.create_socket()
-                    ### self.connect_truesock()
-
             if not self.truesock:
-                raise UnmockedError()
+                raise UnmockedError('Failed to socket.{}() because because a real socket was never created.'.format(function_name))
 
             callback = getattr(self.truesock, function_name)
             return callback(*a, **kw)
@@ -681,23 +737,19 @@ class fakesock(object):
             return self.forward_and_trace('recvfrom_into', *args, **kwargs)
 
         def recv_into(self, *args, **kwargs):
-            if self.truesock and not self.__truesock_is_connected__:
-                self.connect_truesock()
             return self.forward_and_trace('recv_into', *args, **kwargs)
 
         def recvfrom(self, *args, **kwargs):
-            if self.truesock and not self.__truesock_is_connected__:
-                self.connect_truesock()
             return self.forward_and_trace('recvfrom', *args, **kwargs)
 
         def recv(self, buffersize=None, *args, **kwargs):
-            if self.truesock and not self.__truesock_is_connected__:
-                self.connect_truesock()
             buffersize = buffersize or self._bufsize
             return self.forward_and_trace('recv', buffersize, *args, **kwargs)
 
         def __getattr__(self, name):
-            if httpretty.allow_net_connect and not self.truesock:
+            if name in ('getsockopt', ) and not self.truesock:
+                self.truesock = self.create_socket()
+            elif httpretty.allow_net_connect and not self.truesock:
                 # can't call self.connect_truesock() here because we
                 # don't know if user wants to execute server of client
                 # calls (or can they?)
@@ -711,22 +763,34 @@ class fakesock(object):
                         "(see issue https://github.com/gabrielfalcao/HTTPretty/issues/409). "
                         "Please open an issue if this error causes further unexpected issues."
                     )
-                raise UnmockedError()
+
+                raise UnmockedError('Failed to socket.{} because because a real socket does not exist'.format(name))
+
             return getattr(self.truesock, name)
 
+def with_socket_is_secure(sock, kw):
+    sock.is_secure = True
+    sock.kwargs = kw
+    for k, v in kw.items():
+        setattr(sock, k, v)
+    return sock
 
 def fake_wrap_socket(orig_wrap_socket_fn, *args, **kw):
     """drop-in replacement for py:func:`ssl.wrap_socket`
     """
+    if 'sock' in kw:
+        sock = kw['sock']
+    else:
+        sock = args[0]
+
     server_hostname = kw.get('server_hostname')
     if server_hostname is not None:
         matcher = httpretty.match_https_hostname(server_hostname)
         if matcher is None:
-            return orig_wrap_socket_fn(*args, **kw)
-    if 'sock' in kw:
-        return kw['sock']
-    else:
-        return args[0]
+            logger.debug('no requests registered for hostname: "{}"'.format(server_hostname))
+            return with_socket_is_secure(sock, kw)
+
+    return with_socket_is_secure(sock, kw)
 
 
 def create_fake_connection(
@@ -1294,7 +1358,7 @@ class httpretty(HttpBaseClass):
 
     @classmethod
     @contextlib.contextmanager
-    def record(cls, filename, indentation=4, encoding='utf-8'):
+    def record(cls, filename, indentation=4, encoding='utf-8', verbose=False, allow_net_connect=True, pool_manager_params=None):
         """
         .. testcode::
 
@@ -1324,9 +1388,9 @@ class httpretty(HttpBaseClass):
             )
             raise RuntimeError(msg)
 
-        http = urllib3.PoolManager()
+        http = urllib3.PoolManager(**pool_manager_params or {})
 
-        cls.enable()
+        cls.enable(allow_net_connect, verbose=verbose)
         calls = []
 
         def record_request(request, uri, headers):
@@ -1355,7 +1419,7 @@ class httpretty(HttpBaseClass):
                     'headers': dict(response.headers.items())
                 }
             })
-            cls.enable()
+            cls.enable(allow_net_connect, verbose=verbose)
             return response.status, response.headers, response.data
 
         for method in cls.METHODS:
@@ -1368,7 +1432,7 @@ class httpretty(HttpBaseClass):
 
     @classmethod
     @contextlib.contextmanager
-    def playback(cls, filename):
+    def playback(cls, filename, allow_net_connect=True, verbose=False):
         """
         .. testcode::
 
@@ -1386,7 +1450,7 @@ class httpretty(HttpBaseClass):
         :param filename: a string
         :returns: a `context-manager <https://docs.python.org/3/reference/datamodel.html#context-managers>`_
         """
-        cls.enable()
+        cls.enable(allow_net_connect, verbose=verbose)
 
         data = json.loads(open(filename).read())
         for item in data:
@@ -1410,7 +1474,7 @@ class httpretty(HttpBaseClass):
         cls.last_request = HTTPrettyRequestEmpty()
 
     @classmethod
-    def historify_request(cls, headers, body='', append=True):
+    def historify_request(cls, headers, body='', sock=None):
         """appends request to a list for later retrieval
 
         .. testcode::
@@ -1423,12 +1487,16 @@ class httpretty(HttpBaseClass):
 
            assert httpretty.latest_requests[-1].url == 'https://httpbin.org/ip'
         """
-        request = HTTPrettyRequest(headers, body)
+        request = HTTPrettyRequest(headers, body, sock=sock)
         cls.last_request = request
-        if append or not cls.latest_requests:
+
+        if request not in cls.latest_requests:
             cls.latest_requests.append(request)
         else:
-            cls.latest_requests[-1] = request
+            pos = cls.latest_requests.index(request)
+            cls.latest_requests[pos] = request
+
+        logger.info("captured: {}".format(request))
         return request
 
     @classmethod
@@ -1586,16 +1654,18 @@ class httpretty(HttpBaseClass):
         return cls._is_enabled
 
     @classmethod
-    def enable(cls, allow_net_connect=True):
+    def enable(cls, allow_net_connect=True, verbose=False):
         """Enables HTTPretty.
-        When ``allow_net_connect`` is ``False`` any connection to an unregistered uri will throw :py:class:`httpretty.errors.UnmockedError`.
+
+        :param allow_net_connect: boolean to determine if unmatched requests are forwarded to a real network connection OR throw :py:class:`httpretty.errors.UnmockedError`.
+        :param verbose: boolean to set HTTPretty's logging level to DEBUG
 
         .. testcode::
 
            import re, json
            import httpretty
 
-           httpretty.enable()
+           httpretty.enable(allow_net_connect=True, verbose=True)
 
            httpretty.register_uri(
                httpretty.GET,
@@ -1612,9 +1682,13 @@ class httpretty(HttpBaseClass):
 
         .. warning:: after calling this method the original :py:mod:`socket` is replaced with :py:class:`httpretty.core.fakesock`. Make sure to call :py:meth:`~httpretty.disable` after done with your tests or use the :py:class:`httpretty.enabled` as decorator or `context-manager <https://docs.python.org/3/reference/datamodel.html#context-managers>`_
         """
-        cls.allow_net_connect = allow_net_connect
+        httpretty.allow_net_connect = allow_net_connect
         apply_patch_socket()
         cls._is_enabled = True
+        if verbose:
+            logger.setLevel(logging.DEBUG)
+        else:
+            logger.setLevel(logging.getLogger().level)
 
 
 def apply_patch_socket():
@@ -1624,6 +1698,7 @@ def apply_patch_socket():
 
     new_wrap = None
     socket.socket = fakesock.socket
+    socket.socketpair = fake_socketpair
     socket._socketobject = fakesock.socket
     if not bad_socket_shadow:
         socket.SocketType = fakesock.socket
@@ -1634,6 +1709,7 @@ def apply_patch_socket():
     socket.getaddrinfo = fake_getaddrinfo
 
     socket.__dict__['socket'] = fakesock.socket
+    socket.__dict__['socketpair'] = fake_socketpair
     socket.__dict__['_socketobject'] = fakesock.socket
     if not bad_socket_shadow:
         socket.__dict__['SocketType'] = fakesock.socket
@@ -1680,6 +1756,7 @@ def apply_patch_socket():
 
 def undo_patch_socket():
     socket.socket = old_socket
+    socket.socketpair = old_socketpair
     socket.SocketType = old_SocketType
     socket._socketobject = old_socket
 
@@ -1689,6 +1766,7 @@ def undo_patch_socket():
     socket.getaddrinfo = old_getaddrinfo
 
     socket.__dict__['socket'] = old_socket
+    socket.__dict__['socketpair'] = old_socketpair
     socket.__dict__['_socketobject'] = old_socket
     socket.__dict__['SocketType'] = old_SocketType
 
@@ -1746,19 +1824,20 @@ class httprettized(object):
        assert httpretty.latest_requests[-1].url == 'https://httpbin.org/ip'
        assert response.json() == {'origin': '42.42.42.42'}
     """
-    def __init__(self, allow_net_connect=True):
+    def __init__(self, allow_net_connect=True, verbose=False):
         self.allow_net_connect = allow_net_connect
+        self.verbose = verbose
 
     def __enter__(self):
         httpretty.reset()
-        httpretty.enable(allow_net_connect=self.allow_net_connect)
+        httpretty.enable(allow_net_connect=self.allow_net_connect, verbose=self.verbose)
 
     def __exit__(self, exc_type, exc_value, db):
         httpretty.disable()
         httpretty.reset()
 
 
-def httprettified(test=None, allow_net_connect=True):
+def httprettified(test=None, allow_net_connect=True, verbose=False):
     """decorator for test functions
 
     .. tip:: Also available under the alias :py:func:`httpretty.activate`
@@ -1816,7 +1895,7 @@ def httprettified(test=None, allow_net_connect=True):
 
         def new_setUp(self):
             httpretty.reset()
-            httpretty.enable(allow_net_connect)
+            httpretty.enable(allow_net_connect, verbose=verbose)
             if use_addCleanup:
                 self.addCleanup(httpretty.disable)
             if original_setUp:
@@ -1856,7 +1935,6 @@ def httprettified(test=None, allow_net_connect=True):
         except ImportError:
             return False
 
-    "A decorator for tests that use HTTPretty"
     def decorate_class(klass):
         if is_unittest_TestCase(klass):
             return decorate_unittest_TestCase_setUp(klass)
